#version 330 core

out vec4 color;

uniform vec2 windowSize;
uniform mat4 invProjMatr;
uniform mat4 invViewMatr;
uniform vec3 position;

vec2 colorAt(vec3 ray, vec3 pos, vec3 sun, out vec3 newRay, out vec3 newPos, out vec4 col);
vec2 traceRay(vec3 ray, vec3 pos, int generation, int excludeIndex);

const vec4[] spheres = {
	vec4(2, 1, -4, 1),//x, y, z, r
	vec4(-1, -1.5, -6, 1.5),
	vec4(-1.5, 0, -4.5, 0.5)
};

void main() {
	vec4 coordsBeforeMatr = vec4(gl_FragCoord.x / windowSize.x * 2 - 1, gl_FragCoord.y / windowSize.y * 2 - 1, -1, 1);
	vec4 coordsAfterProj = invProjMatr * coordsBeforeMatr;
	vec3 ray = normalize((invViewMatr * vec4(coordsAfterProj.xy, -1, 0)).xyz);
	vec3 sun = normalize(vec3(1, 2, 3));

	vec3 pos = position;
	int lastIndex = -1;
	const int iterations = 10;
	float colorMult = 1;
	float colorMultAccum = 0;
	for (int i = 0; i < iterations; i++) {
		vec4 col = vec4(0, 0, 0, 1);
		vec3 newPos, newRay;

		vec2 results = traceRay(ray, pos, i, lastIndex);
		int closestIndex = int(results.x);
		float closestDist = results.y;

		if (closestIndex != -1) {
			newPos = pos + ray * closestDist;
			vec2 res2 = traceRay(sun, newPos, i+1, closestIndex);
			vec3 diff = spheres[closestIndex].xyz - pos;
			vec3 normal = normalize(ray * closestDist - diff);
			newRay = reflect(ray, normal);
			if (res2.x < -0.5) {
				float brightness = max(dot(normal, sun), 0);
				col = vec4(mod(0.33 * closestIndex, 1), mod(0.33 * (closestIndex + 1), 1), mod(0.33 * (closestIndex + 2), 1), 0.0) * brightness;
				col.a = 1;
			}
		} else if (dot(sun, ray) > 0.9999) {
			col = vec4(vec3(1, 1, 1) * cos(pow(acos(dot(sun, ray)) / acos(0.9999), 2) * 3.14159265358979 / 2), 1);
			i = iterations;//break the loop
		} else i = iterations;

		lastIndex = closestIndex;
		ray = newRay;
		pos = newPos;

		color += colorMult * col;
		colorMult = colorMult * 0.9;
		colorMultAccum += colorMult;
	}
}

vec2 traceRay(vec3 ray, vec3 pos, int generation, int excludeIndex) {
	int closestIndex = -1;
	float closestDist = 1e9;
	for (int i = 0; i < 3; i++) {
		if (i == excludeIndex) continue;

		vec3 diff = spheres[i].xyz - pos;
		vec3 proj = dot(diff, ray) * ray;

		float a = dot(ray, diff);
		float b = a * a - dot(diff, diff) + spheres[i].w * spheres[i].w;

		if (b >= 0) {//Sphere found!!
			float rt = sqrt(b);
			float d1 = a - rt;
			float d2 = a + rt;
			if (d1 > 0) {
				if (d1 < closestDist) {
					closestDist = d1;
					closestIndex = i;
				}
			} else if (d2 > 0) {
				if (d2 < closestDist) {
					closestDist = d2;
					closestIndex = i;
				}
			}
		}
	}

	return vec2(closestIndex, closestDist);
}
